"""
Module with funtionality to collect and calculate system metrics
"""
from statistics import mean
from time import sleep

from . import si
from .interval import Interval
from .utils import noop

def avgs(items, key_blacklist=()):
    """
    Expects a list of equally structured dicts and returns a dict with the mean
    of all values. For blacklisted keys the last value is used.

    Example:
    ```python
    in = [{ a: 1, b: 10, c: 1 }, { a: 2, b: 20, c: 2 }, { a: 3, b: 30, c: 3 }]
    avgs(in, ('c'))
    >>> { a: 2, b: 20, c: 3 }
    ```
    """
    keys = items[0].keys()
    values = dict()
    # Initialize all keys with empty lists so we can append
    for key in keys:
        values[key] = []
    # Append all values per key
    for item in items:
        for key in keys:
            values[key].append(item.get(key))
    # Calculate averages for keys not in `key_blacklist`, otherwise take last value
    result = dict()
    for key in keys:
        result[key] = round(mean(values[key]), 3) if not key in key_blacklist else values[key][-1]
    return result

def poll():
    """
    Polls and returns system information
    """
    data = dict()

    data.update(host=si.host())
    data.update(ts=si.timestamp())
    loadavgs = si.loadavg()
    data.update(loadavg_1min=loadavgs[0])
    data.update(loadavg_5min=loadavgs[1])
    data.update(loadavg_15min=loadavgs[2])
    data.update(mem=si.memory())
    data.update(cpu=si.cpu())
    data.update(disk=si.disk())
    diskio = si.diskio()
    data.update(diskio_rps=diskio[0])
    data.update(diskio_wps=diskio[1])
    netio = si.netio()
    data.update(netio_rps=netio[0])
    data.update(netio_sps=netio[1])

    return data

class Collector:
    """
    Collects the system metrics every second (tick) and passes them to
    `on_input` every `loginterval` seconds (log). By default the log interval is
    also one second, if it's above, averages will be passed.
    Additionally it also accepts `on_tick` as an additional and optional
    callback for every tick.
    """
    NO_AVG_KEYS = ('host', 'ts') # Keys of which no average is calculated for

    @staticmethod
    def create(*args):
        """
        Static constructor
        """
        return Collector(*args)

    def __init__(self, on_input, loginterval=1, on_tick=noop):
        if not callable(on_input):
            raise TypeError('on_input must be callable')
        if not isinstance(loginterval, int) or loginterval < 1:
            raise ValueError('loginterval must be an integer equal or above 1')
        if on_tick is not noop and not callable(on_tick):
            raise TypeError('on_tick must be callable if provided')

        self.loginterval = loginterval
        self.on_input = on_input
        self.on_tick = on_tick
        self.tick_values = []

        # Initial poll and delay, so next poll has IO data
        poll()
        sleep(1)
        self.tick_index = 1

        Interval.create(self.collect, 1)

    def collect(self):
        """
        Collects data on each tick and passes them to the callbacks if appropriate
        """
        data = poll()
        self.on_tick(data)
        if self.loginterval == 1:
            self.on_input(data)
        elif self.tick_index % self.loginterval == 0:
            self.tick_values.append(data)
            self.on_input(avgs(self.tick_values, self.NO_AVG_KEYS))
            self.tick_values.clear()
        else:
            self.tick_values.append(data)
        self.tick_index += 1



"""
Provides funtionalities for time intervals
"""
import sched
import time

class Interval:
    """
    Sets up an interval calling `fn` every time `interval` passes
    """
    def __init__(self, fn, interval, autostart=True):
        self._fn = fn
        self._interval = interval
        self._scheduler = sched.scheduler(time.time, time.sleep)
        self._ev = None
        if autostart:
            self.start()

    @staticmethod
    def create(*args):
        """
        Static constructor
        """
        return Interval(*args)

    def _call(self):
        self._fn()
        self._next()

    def _next(self):
        self._ev = self._scheduler.enter(self._interval, 1, self._call)

    def start(self):
        """
        Starts the interval, only relevant if autostart=False on construction
        """
        self._next()
        self._scheduler.run()



from os import path

from setuptools import setup, find_packages

import simond

def long_descr():
    with open(
        path.join(path.abspath(path.dirname(__file__)), 'README.md'),
        encoding='utf-8'
    ) as f:
        return f.read()

setup(
    name='simond',
    version=simond.__version__,
    description=simond.__doc__.strip(),
    license=simond.__license__,
    long_description=long_descr(),
    long_description_content_type='text/markdown',
    author=simond.__author__,
    author_email='ronnyhaase@fastmail.com',
    url='https://github.com/ronnyhaase/simon',
    project_urls={
        "Bug Tracker": "https://github.com/ronnyhaase/simon/issues",
        "Documentation": "https://github.com/ronnyhaase/simon/daemon",
        "Source Code": "https://github.com/ronnyhaase/simon/daemon",
    },
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Topic :: System :: Monitoring"
    ],
    python_requires='>=3.6',
    packages=find_packages(),
    install_requires=[
        'psutil>=5.7,<6',
        'setuptools'
    ],
    extras_require={
        'dev': [
            'pylint'
        ]
    },
    entry_points={
        'console_scripts': [
            'simond = simond.__main__:main'
        ]
    }
)



"""
Provides functionality to gather system information
"""
import platform
import datetime

import psutil

def cpu():
    """
    Gathers and returns total CPU usage in percent as decimal
    """
    sin = psutil.cpu_percent()
    return round(sin / 100, 3)

def disk():
    """
    Gathers and returns total disk usage in percent as decimal
    """
    sin = psutil.disk_usage('/')
    return round(sin.percent / 100, 3)

def _init_diskio():
    prev = None
    def update():
        nonlocal prev
        result = [None, None]
        curr = psutil.disk_io_counters()
        if prev:
            result = [
                curr.read_bytes - prev.read_bytes,
                curr.write_bytes - prev.write_bytes
            ]
        prev = curr
        return result
    return update
diskio = _init_diskio()
diskio.__doc__ = """
    Returns the disk IO since the last call in bytes read and written
"""

def host():
    """
    Gathers and returns the host's name
    """
    return platform.node()

def loadavg():
    """
    Gathers and returns the system's load averages
    """
    sin = psutil.getloadavg()
    return [
        round(sin[0], 3),
        round(sin[1], 3),
        round(sin[2], 3)
    ]

def memory():
    """
    Gathers and returns the total memory usage in percent as decimal
    """
    sin = psutil.virtual_memory()
    return round((sin.total / sin.used) / 100, 3)

def _init_netio():
    prev = None
    def update():
        nonlocal prev
        result = [None, None]
        curr = psutil.net_io_counters()
        if prev:
            result = [
                curr.bytes_recv - prev.bytes_recv,
                curr.bytes_sent - prev.bytes_sent
            ]
        prev = curr
        return result
    return update
netio = _init_netio()
netio.__doc__ = """
     Returns the network IO since the last call in bytes received and sent
"""

def timestamp():
    """
    Returns current UTC UNIX time
    """
    return round(datetime.datetime.utcnow().timestamp())



"""
simond entry point
"""
from argparse import ArgumentParser #, FileType
import json
import sys

from . import __version__, __doc__
from .collector import Collector

def parse_args():
    """
    Defines, handles and returns program args
    """
    apa = ArgumentParser('simond', description=__doc__.strip())
    # apa.add_argument(
    #     '-o',
    #     '--output',
    #     type=FileType('a', 1),
    #     required=True,
    #     help='The file where to log system information to'
    # )
    apa.add_argument(
        '-V',
        '--version',
        help='Print simond version number',
        action='store_true'
    )
    apa.add_argument(
        '-i',
        '--loginterval',
        type=int,
        required=False,
        default=1,
        help="The log interval in seconds, if it's anything but 1, averages \
            (sample mean) will be logged"
    )
    return apa.parse_args()

def main():
    """
    main
    """
    args = parse_args()

    if args.version:
        print(__version__)
        sys.exit(0)

    def log(data):
        # args.output.write(json.dumps(data) + "\n")
        print(json.dumps(data))

    try:
        Collector.create(log, args.loginterval)
    except KeyboardInterrupt:
        pass
    except SystemExit:
        pass
    # finally:
    #     args.output.close()

if __name__ == '__main__':
    main()
